/*
    Copyright 2007-2017 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

/*
How to compile:
windres winamp.rc rc.o
gcc -o wampmini.exe wampmini.c mylibaow32.c rc.o -lwinmm -O2 -s
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <windows.h>
#include <fcntl.h>
#include "in2.h"
#include "mywav.h"
#include "ao.h"



#define VER             "0.1.3"
#define TITLE           "Mini Winamp input plugins player/converter " VER
#define WM_WA_MPEG_EOF  WM_USER+2



FILE    *fd             = NULL;
int     play            = 0;    // 1:audio, -1:stdout, 0:file
char    *outwav         = NULL;
mywav_fmtchunk  wavfmt;
ao_device       *device = NULL;
char    outfile[2000]   = "";



void std_err(char *err) {
    if(err) {
        fprintf(stderr, "\nError: %s\n", err);
    } else {
        perror("\nError");
    }
    exit(1);
}



int Open(int samplerate, int numchannels, int bitspersamp, int bufferlenms, int prebufferms) {
    ao_sample_format    format;

    fprintf(stderr, "- format %d %d %d\n", samplerate, numchannels, bitspersamp);

    if(play > 0) {
        fprintf(stderr, "- output: audio\n");
        ao_initialize();
        format.bits        = bitspersamp;
        format.channels    = numchannels;
        format.rate        = samplerate;
        format.byte_format = AO_FMT_LITTLE;
        device = ao_open_live(ao_default_driver_id(), &format, NULL);
        if(!device) std_err("aolib");
        return 0;

    } else if(play < 0) {
        fprintf(stderr, "- output: stdout\n");
        fd = stdout;
        // stdout must have a RIFF header because channels/samplerate/bits are not standard

    } else {
        fprintf(stderr, "- output %s\n", outfile);
        fd = fopen(outfile, "wb");
        if(!fd) std_err(NULL);
    }

    wavfmt.wFormatTag       = 0x0001;
    wavfmt.wChannels        = numchannels;
    wavfmt.dwSamplesPerSec  = samplerate;
    wavfmt.wBitsPerSample   = bitspersamp;
    wavfmt.wBlockAlign      = (wavfmt.wBitsPerSample / 8) * wavfmt.wChannels;
    wavfmt.dwAvgBytesPerSec = wavfmt.dwSamplesPerSec * wavfmt.wBlockAlign;

    mywav_writehead(fd, &wavfmt, 0x7fffff00 /*hopefully is better for stdout if enabled and works with bugged encoders too*/, NULL, 0);
    return 0;
}



int Write(char *buf, int len) {
    if(play > 0) {
        ao_play(device, buf, len);
    } else {
        fwrite(buf, 1, len, fd);
    }
    return 0;
}



void Close(void) {
    int     fsize;

    if(play > 0) {
        Sleep(1000);
        ao_close(device);
        return;
    }

    if(fd == stdout) return;

    fsize = (int)ftell(fd);
    fseek(fd, 0, SEEK_SET);
    mywav_writehead(fd, &wavfmt, fsize - 44, NULL, 0);
    fprintf(stderr, "- %d bytes written\n", fsize);
    fclose(fd);
}



int CanWrite(void) { return INTPTR_MAX; }
int IsPlaying(void) { return 0; }
int Pause(int pause) { return !pause; }
void Config(HWND hwndParent) {}
void About(HWND hwndParent) {}
void Init(void) {}
void Quit(void) {}
void SetVolume(int volume) {}
void SetPan(int pan) {}
void Flush(int t) {}
int GetOutputTime(void) { return 0; }
int GetWrittenTime(void) { return 0; }
void SAVSAInit(int maxlatency_in_ms, int srate) {}
void SAVSADeInit(void) {}
void SAAddPCMData(void *PCMData, int nch, int bps, int timestamp) {}
int SAGetMode(void) { return 0; }
int SAAdd(void *data, int timestamp, int csa) { return 0; }
void VSAAddPCMData(void *PCMData, int nch, int bps, int timestamp) {}
int VSAGetMode(int *specNch, int *waveNch) { return 0; }
int VSAAdd(void *data, int timestamp) { return 0; }
void VSASetInfo(int nch, int srate) {}
int dsp_isactive(void) { return 0; }
int dsp_dosamples(short int *samples, int numsamples, int bps, int nch, int srate) { return numsamples; }
void EQSet(int on, char data[10], int preamp) {}
void SetInfo(int bitrate, int srate, int stereo, int synched) {}



LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam) {
    switch(msg) {
        case WM_WA_MPEG_EOF:    DestroyWindow(hwnd);    break;
        case WM_CLOSE:          DestroyWindow(hwnd);    break;
        case WM_DESTROY:        PostQuitMessage(0);     break;
        default:                return DefWindowProc(hwnd, msg, wParam, lParam);    break;
    }
    return 0;
}



HWND create_hwnd(void) {
    WNDCLASSEX  wc;
    HWND        hwnd;
    static const char classname[] = "wampmini";

    memset(&wc, 0, sizeof(wc));
    wc.cbSize        = sizeof(WNDCLASSEX);
    wc.lpfnWndProc   = WndProc;
    wc.lpszClassName = classname;
    if(!RegisterClassEx(&wc)) std_err("RegisterClassEx");

    hwnd = CreateWindowEx(
        WS_EX_CLIENTEDGE,
        classname,
        TITLE,
        WS_DISABLED,
        CW_USEDEFAULT, CW_USEDEFAULT, 0, 0,
        NULL, NULL, NULL, NULL);
    if(!hwnd) std_err("CreateWindowEx");

    return hwnd;
}



int main(int argc, char *argv[]) {
    In_Module   *(WINAPI *winampGetInModule2)();
    HINSTANCE   hlib;
    Out_Module  outm;
    In_Module   *inm;
    HWND        hwnd;
    MSG         Msg;
    char        *indll,
                infile[2000],
                *np;

    setbuf(stdin,  NULL);
    setbuf(stdout, NULL);
    setbuf(stderr, NULL);

    #ifdef O_BINARY
    setmode(fileno(stdin), O_BINARY);
    setmode(fileno(stdout), O_BINARY);
    #endif

    fputs("\n"
        TITLE "\n"
        "by Luigi Auriemma\n"
        "e-mail: me@aluigi.org\n"
        "web:    aluigi.org\n"
        "\n", stderr);

    if(argc < 4) {
        fprintf(stderr, "\n"
            "Usage: %s <in_plugin.DLL> <input> <output*>\n"
            "\n"
            "Output can be:\n"
            "- a file like output.wav\n"
            "- stdout using -\n"
            "- Windows audio using PLAY or play\n"
            "\n", argv[0]);
        exit(1);
    }
    indll  = argv[1];
    infile[GetFullPathName(argv[2], sizeof(infile) - 1, infile, &np)] = 0;
    outwav = argv[3];
    if(!stricmp(outwav, "PLAY")) {
        play = 1;
    } else if(!stricmp(outwav, "-")) {
        play = -1;
    } else {
        outfile[GetFullPathName(outwav, sizeof(outfile) - 1, outfile, &np)] = 0;
    }

    // win9x compatibility, why? because it's possible but totally useless because the plugins don't load or work
    static HMODULE WINAPI (*_LoadLibraryExA)(LPCTSTR lpFileName, HANDLE  hFile, DWORD dwFlags) = NULL;
    HANDLE  kernel32;
    kernel32 = GetModuleHandle("kernel32.dll");
    if(kernel32) _LoadLibraryExA = (void *)GetProcAddress(kernel32, "LoadLibraryExA");

    // used by various plugins, no need to check if it exists (in my tests it wasn't useful)
    //LoadLibrary("msvcr90.dll");   // don't load it or in_vgmstream may display an error before working

    hlib = NULL;
    if(_LoadLibraryExA) {
        if(!hlib) hlib = _LoadLibraryExA(indll, NULL, /*LOAD_LIBRARY_SEARCH_DLL_LOAD_DIR*/ 0x00000100); // needed for in_vgmstream
        if(!hlib) hlib = _LoadLibraryExA(indll, NULL, /*LOAD_LIBRARY_SEARCH_DEFAULT_DIRS*/ 0x00001000);
    }
    if(!hlib) hlib = LoadLibrary(indll);
    if(_LoadLibraryExA) {
        if(!hlib) hlib = _LoadLibraryExA(indll, NULL, /*DONT_RESOLVE_DLL_REFERENCES*/ 0x00000001);  // last hope?
    }
    if(!hlib) std_err("dll not found or missing dependencies");
    fprintf(stderr, "- hlib %p\n", hlib);

    winampGetInModule2 = (void *)GetProcAddress(hlib, "winampGetInModule2");
    if(!winampGetInModule2) {
        FreeLibrary(hlib);
        std_err("function winampGetInModule2 not found");
    }
    fprintf(stderr, "- winampGetInModule2 %p\n", winampGetInModule2);

    hwnd = create_hwnd();

    memset(&outm, 0, sizeof(outm)); // useless
    outm.version        = OUT_VER;
    outm.description    = TITLE;
    outm.id             = 65536;
    outm.hMainWindow    = hwnd;
    outm.hDllInstance   = hlib;
    outm.Config         = Config;
    outm.About          = About;
    outm.Init           = Init;
    outm.Quit           = Quit;
    outm.Open           = Open;
    outm.Close          = Close;
    outm.Write          = Write;
    outm.CanWrite       = CanWrite;
    outm.IsPlaying      = IsPlaying;
    outm.Pause          = Pause;
    outm.SetVolume      = SetVolume;
    outm.SetPan         = SetPan;
    outm.Flush          = Flush;
    outm.GetOutputTime  = GetOutputTime;
    outm.GetWrittenTime = GetWrittenTime;

    inm                 = winampGetInModule2();
    inm->outMod         = &outm;
    inm->hMainWindow    = hwnd;
    inm->hDllInstance   = hlib;
    inm->SAVSAInit      = SAVSAInit;
    inm->SAVSADeInit    = SAVSADeInit;
    inm->SAAddPCMData   = SAAddPCMData;
    inm->SAGetMode      = SAGetMode;
    inm->SAAdd          = SAAdd;
    inm->VSAAddPCMData  = VSAAddPCMData;
    inm->VSAGetMode     = VSAGetMode;
    inm->VSAAdd         = VSAAdd;
    inm->VSASetInfo     = VSASetInfo;
    inm->dsp_isactive   = dsp_isactive;
    inm->dsp_dosamples  = dsp_dosamples;
    inm->EQSet          = EQSet;
    inm->SetInfo        = SetInfo;

    if(inm->description)    fprintf(stderr, "- description: %s\n", inm->description);
    //if(inm->FileExtensions) fprintf(stderr, "- FileExtensions: %s\n", inm->FileExtensions);

    inm->Init();
    //if(inm->IsOurFile) inm->IsOurFile(infile);
    inm->Play(infile);

    while(GetMessage(&Msg, NULL, 0, 0) > 0) {
        TranslateMessage(&Msg);
        DispatchMessage(&Msg);
    }

    inm->Stop();
    inm->Quit();
    FreeLibrary(hlib);

    fprintf(stderr, "- finished\n");
    return 0;
}


